import std/strutils
import zippy


type
  Dictionary* = object
    words: string
    offsets: seq[uint32]


proc `[]`*(d: Dictionary, i: Natural): string =
  # last word has no following start index, so we have to fake it
  # also strings are indexed with ints
  let slice = 
    if i == d.offsets.high:
      d.offsets[i].int .. d.words.high
    else:
      d.offsets[i].int ..< d.offsets[i + 1].int

  result = d.words[slice]


proc len*(d: Dictionary): int =
  result = d.offsets.len


proc addU32(s: var seq[uint8], i: uint32) =
  for offset in 0..3:
    let b = cast[uint8](i shr (offset * 8))
    s.add(b)


proc getU32(s: seq[uint8]): uint32 =
  for offset in 0..3:
    result = result or (cast[uint32](s[offset]) shl (offset * 8))


proc pack*(d: Dictionary): seq[uint8] =
  let compressed = cast[seq[uint8]](d.words)
  let lenWords = compressed.len
  let lenOffsets = d.offsets.len * 4
  # 8 extra bytes for the length specifiers
  var buff = newSeqOfCap[uint8](lenWords + lenOffsets + 8)
  buff.addU32(lenWords.uint32)
  buff.add(compressed)
  buff.addU32(lenOffsets.uint32)
  for offset in d.offsets:
    buff.addU32(offset)
  result = buff.compress(dataFormat = dfGzip)


proc unpack*[T: seq[uint8]|string](p: T): Dictionary =
  when T is string:
    let buff = cast[seq[uint8]](p.uncompress(dataFormat = dfGzip))
  else:
    let buff = p.uncompress(dataFormat = dfGzip)
  
  let lenWords = buff.getU32()
  let nextSection = lenWords + 4
  let words = cast[string](buff[4 ..< nextSection])

  let numOffsets = buff[nextSection ..< (nextSection + 4)].getU32() div 4
  var offsets = newSeqOfCap[uint32](numOffsets)
  for i in 0 ..< numOffsets:
    let idx = nextSection + 4 + (i * 4)
    let offset = buff[idx ..< idx + 4].getU32()
    offsets.add(offset)

  result = Dictionary(words: words, offsets: offsets)


proc loadWords*(): Dictionary =
  var i: uint32
  for word in readFile("../data/dictionary.txt").strip().splitLines():
    result.words.add(word)
    result.offsets.add(i)
    i += word.len.uint32


when isMainModule:
  echo "Loading words..."
  let dictionary = loadWords()
  echo "Packing dictionary..."
  let packed = dictionary.pack()
  discard open("../data/dictionary.pack", fmWrite).writeBytes(packed, 0, packed.len)
  echo "Dictionary packed."
